==> blcr-0.2.0.tar.gz: blcr-0.2.0/libcr/cr_cs.c <==
/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_cs.c,v 1.44 2003/11/14 19:55:24 phargrov Exp $
 */

#include <errno.h>

#include "cr_private.h"

//
// Private data
//

#define CRI_CONTEXT_MASK    (CR_THREAD_CONTEXT | CR_SIGNAL_CONTEXT)

//
// Private functions
//

static inline void
poll_checkpoint(cri_info_t *info)
{
    if (cri_atomic_dec_and_test(&info->cr_cs_count)) {
	cri_black_lock(&cri_cs_lock);
	cri_start_checkpoint(info, 0);
	cri_black_unlock(&cri_cs_lock);
    }
}

// signal_handler()
//
// Signal handler for threads w/ signal-context handlers
static void
my_handler(int signr, siginfo_t *siginfo, void *context)
{
    int saved_errno = errno;
    cri_info_t *info = cri_info_location();	// thread-specific

    // Process argument(s) from the kernel
    cri_get_siginfo(info, siginfo);

    if (cri_cmp_swap(&info->cr_state, CR_STATE_IDLE, CR_STATE_PENDING)) {
	cri_atomic_inc(&cri_live_count);
	poll_checkpoint(info);
    }

    errno = saved_errno;
}

static int
do_state(cri_info_t	*info)
{
    int state = cri_atomic_read(&info->cr_state);

    if ((state == CR_STATE_IDLE) && cri_atomic_read(&cri_live_count)) {
	state = CR_STATE_PENDING;
    }

    return state;
}

// XXX: Unused 'id' argument should be used for debugging.
// XXX: we could also double check that (id == CR_ID_CALLBACK)
// if and only if the state is CR_STATE_ACTIVE.
//
// Returns the current state.
//
// Signal safe because if a signal takes us into the ACTIVE state
// it will also leave the ACTIVE state before we regain control.
//
static int
do_enter(cri_info_t	*info,
	 cr_client_id_t	id)
{
    int state;

    state = cri_atomic_read(&info->cr_state);

    if (state != CR_STATE_ACTIVE) {
	cri_atomic_t *count = &info->cr_cs_count;
	cri_atomic_t old;

	// This mess is to prevent starvation by not entering an
	// outermost critical section when a checkpoint is pending
	// or running in ANY thread.
	do {
	    // Note that we read from 'count' first to avoid later using
	    // an uninitialized value for 'old' in the cri_cmp_swap().
	    // Also note that while we should NEVER observe a case of
	    // (cr_cs_count == 0) here, we test w/ <= 1 just in case.
	    CRI_WHILE_COND_YIELD(((old = cri_atomic_read(count)) <= 1) &&
				 cri_atomic_read(&cri_live_count));
	} while (!cri_cmp_swap(count, old, old + 1));

	cri_red_lock(&cri_cs_lock);
    }

    return state;
}
	 
// XXX: Unused 'id' argument should be used for debugging.
static void
do_leave(cri_info_t	*info,
	 cr_client_id_t	id)
{
    if (cri_atomic_read(&info->cr_state) != CR_STATE_ACTIVE) {
	cri_red_unlock(&cri_cs_lock);
	poll_checkpoint(info);
    }
}

// Not reentrant.
static cr_callback_id_t
cri_register_signal(cri_info_t*		info,
		    cr_callback_t	func,
		    void*		arg,
		    int			flags)
{
    return cri_do_register(info, func, arg, flags);
}

// Not reentrant.
static int
cri_replace_signal(cri_info_t		*info,
		   cr_callback_id_t	id,
		   cr_callback_t	func,
		   void*		arg,
		   int			flags)
{
    int retval = -1;	// assume failure
    int state;

    state = do_enter(info, id);

    if (state != CR_STATE_ACTIVE) {
	retval = cri_do_replace(info, id, func, arg, flags);
    } else {
	// Called from a callback!  retval is already -1;
    }

    do_leave(info, id);

    return retval;
}

//
// "Public" functions
//

// Not reentrant.
cr_callback_id_t
cri_do_register(cri_info_t	*info,
	        cr_callback_t	func,
	        void*		arg,
	        int		flags)
{
    int indx = info->cr_cb_count;

    if (indx < CR_MAX_CALLBACKS) {
	int next = indx + 1;

	// Set entry before advancing the count
	info->cr_cb[indx].func  = func;
	info->cr_cb[indx].arg   = arg;
	info->cr_cb[indx].flags = flags;
	info->cr_cb_count = next;

	return indx;
    } else {
	return -1;
    }
}

// Note that we don't replace 'func', 'arg' and 'flags' atomically.
// Note that we ensure the context is not changed.
// Such atomicity is the responsibility of the caller.
//
// Not reentrant.
// Not exposed to user.
int
cri_do_replace(cri_info_t	*info,
	       cr_callback_id_t	id,
	       cr_callback_t	func,
	       void*		arg,
	       int		flags)
{
    int retval = -1;

    if ((flags & CRI_CONTEXT_MASK) == (id & CRI_CONTEXT_MASK)) {
	id &= ~CRI_CONTEXT_MASK;
	if ((id >= 0) && (id < info->cr_cb_count)) {
	    info->cr_cb[id].func  = func;
	    info->cr_cb[id].arg   = arg;
	    info->cr_cb[id].flags = flags;
	    retval = 0;
	} else {
	    /* id out-of-bounds */
	}
    } else {
	/* Context mismatch */
    }

    return retval;
}

cr_callback_id_t
cr_register_callback(cr_callback_t	func,
		     void*		arg,
		     int		flags)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    int context;
    int retval;

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    context = flags & CRI_CONTEXT_MASK;

    switch (context) {
    case CR_SIGNAL_CONTEXT:
	retval = cri_register_signal(info, func, arg, flags);
	break;
	
    case CR_THREAD_CONTEXT:
	retval = cri_register_thread(info, func, arg, flags);
	break;

    default:
	retval = -1;
    }

    return (retval >= 0) ? (retval | context) : retval;
}

// Not reentrant.
int
cr_replace_callback(cr_callback_id_t	id,
		    cr_callback_t	func,
		    void*		arg,
		    int			flags)
{
    cri_info_t *info = cri_info_location();    // thread-specific
    int retval;

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    switch (flags & CRI_CONTEXT_MASK) {
    case CR_SIGNAL_CONTEXT:
	retval = cri_replace_signal(info, id, func, arg, flags);
	break;
	
    case CR_THREAD_CONTEXT:
	retval = cri_replace_thread(info, id, func, arg, flags);
	break;

    default:
	retval = -1;
    }

    return retval;
}

// Not reentrant.
int
cr_replace_self(cr_callback_t	func,
		void*		arg,
		int		flags)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    cr_callback_id_t id;

    if (!info || (cri_atomic_read(&info->cr_state) != CR_STATE_ACTIVE)) {
	// Not called from a callback!
	return -1;
    }

    /* form the full id from the index and context */
    id = info->run.index |
		(info->cr_cb[info->run.index].flags & CRI_CONTEXT_MASK);

    return cri_do_replace(info, id, func, arg, flags);
}

int
cr_status(void)
{
    cri_info_t *info = cri_info_location();	// thread-specific

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    return do_state(info);
}

// XXX: Unused 'id' argument should be used for debugging.
void
cr_enter_cs(cr_client_id_t id)
{
    cri_info_t *info = cri_info_location();	// thread-specific

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    do_enter(info, id);
}

// XXX: Unused 'id' argument should be used for debugging.
void
cr_leave_cs(cr_client_id_t id)
{
    cri_info_t *info = cri_info_location();	// thread-specific

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    do_leave(info, id);
}

// Return address of the per-checkpoint/per-thread info
const struct cr_checkpoint_info * cr_get_checkpoint_info(void)
{
    cri_info_t *info = cri_info_location();		// thread-specific
    return &(info->cr_checkpoint_info);
}

// Return a client identifier used to invoke cr_{enter,leave}_cs().
// Also registers the handler on first call per thread.
// May later setup additional state.
cr_client_id_t cr_init(void)
{
    cri_info_t *info;
    int rc;

    // First verify that we have kernel support
    rc = cri_get_fd();
    if (rc < 0) {
	return rc;
    }

    info = cri_info_init();		// thread-specific
    if (!info->handler) {
	// Set the handler
	info->register_cmd = CR_OP_HAND_PHASE2; // MUST WRITE FIRST
	info->handler = &my_handler;

	// Register ourself with the kernel
	do {
	    rc = cri_syscall(CR_OP_HAND_PHASE2, &my_handler);
	} while ((rc < 0) && (errno == EAGAIN));
	if (rc != 0) {
#if 1
	    // XXX: while rerestart-time re-registration is done in user
	    // space we must check for EEXIST due to a potential race
	    // where we are "re-registered" before the first registration.
	    if (errno != EEXIST) {
		CRI_ABORT("CR_OP_HAND_PHASE2 returned w/ errno = %d", errno);
	    }
#else
	    CRI_ABORT("CR_OP_HAND_PHASE2 returned w/ errno = %d", errno);
#endif
	}
    }

    return info->next_id++;
}

==> blcr-0.5.0.tar.gz: blcr-0.5.0/libcr/cr_cs.c <==
/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_cs.c,v 1.51 2007/01/05 23:47:38 phargrov Exp $
 */

#include <errno.h>

#include "cr_private.h"

//
// Private data
//

#define CRI_CONTEXT_MASK    (CR_THREAD_CONTEXT | CR_SIGNAL_CONTEXT)

//
// Private functions
//

static inline void
poll_checkpoint(cri_info_t *info)
{
    if (cri_atomic_dec_and_test(&info->cr_cs_count)) {
	cri_black_lock(&cri_cs_lock);
	cri_start_checkpoint(info, 0);
	cri_black_unlock(&cri_cs_lock);
    }
}

// signal_handler()
//
// Signal handler for threads w/ signal-context handlers
static void
my_handler(int signr, siginfo_t *siginfo, void *context)
{
    int saved_errno = errno;
    cri_info_t *info = cri_info_location();	// thread-specific

    // Process argument(s) from the kernel
    cri_checkpoint_info_init(info, siginfo);

    if (cri_cmp_swap(&info->cr_state, CR_STATE_IDLE, CR_STATE_PENDING)) {
	cri_atomic_inc(&cri_live_count);
	poll_checkpoint(info);
    }

    errno = saved_errno;
}

static int
do_state(cri_info_t	*info)
{
    int state = cri_atomic_read(&info->cr_state);

    if ((state == CR_STATE_IDLE) && cri_atomic_read(&cri_live_count)) {
	state = CR_STATE_PENDING;
    }

    return state;
}

// XXX: Unused 'id' argument should be used for debugging.
// XXX: we could also double check that (id == CR_ID_CALLBACK)
// if and only if the state is CR_STATE_ACTIVE.
//
// Returns the current state.
//
// Signal safe because if a signal takes us into the ACTIVE state
// it will also leave the ACTIVE state before we regain control.
//
static int
do_enter(cri_info_t	*info,
	 cr_client_id_t	id)
{
    int state;

    state = cri_atomic_read(&info->cr_state);

    if (state != CR_STATE_ACTIVE) {
	cri_atomic_t *count = &info->cr_cs_count;
	cri_atomic_t old;

	// This mess is to prevent starvation by not entering an
	// outermost critical section when a checkpoint is pending
	// or running in ANY thread.
	do {
	    // Note that we read from 'count' first to avoid later using
	    // an uninitialized value for 'old' in the cri_cmp_swap().
	    // Also note that while we should NEVER observe a case of
	    // (cr_cs_count == 0) here, we test w/ <= 1 just in case.
	    CRI_WHILE_COND_YIELD(((old = cri_atomic_read(count)) <= 1) &&
				 cri_atomic_read(&cri_live_count));
	} while (!cri_cmp_swap(count, old, old + 1));

	cri_red_lock(&cri_cs_lock);
    }

    return state;
}
	 
// XXX: Unused 'id' argument should be used for debugging.
static void
do_leave(cri_info_t	*info,
	 cr_client_id_t	id)
{
    if (cri_atomic_read(&info->cr_state) != CR_STATE_ACTIVE) {
	cri_red_unlock(&cri_cs_lock);
	poll_checkpoint(info);
    }
}

// Not reentrant.
static cr_callback_id_t
cri_register_signal(cri_info_t*		info,
		    cr_callback_t	func,
		    void*		arg,
		    int			flags)
{
    return cri_do_register(info, func, arg, flags);
}

// Not reentrant.
static int
cri_replace_signal(cri_info_t		*info,
		   cr_callback_id_t	id,
		   cr_callback_t	func,
		   void*		arg,
		   int			flags)
{
    int retval = -1;	// assume failure
    int state;

    state = do_enter(info, id);

    if (state != CR_STATE_ACTIVE) {
	retval = cri_do_replace(info, id, func, arg, flags);
    } else {
	// Called from a callback!  retval is already -1;
    }

    do_leave(info, id);

    return retval;
}

//
// "Public" functions
//

// Not reentrant.
cr_callback_id_t
cri_do_register(cri_info_t	*info,
	        cr_callback_t	func,
	        void*		arg,
	        int		flags)
{
    int indx = info->cr_cb_count;

    if (indx < CR_MAX_CALLBACKS) {
	int next = indx + 1;

	// Set entry before advancing the count
	info->cr_cb[indx].func  = func;
	info->cr_cb[indx].arg   = arg;
	info->cr_cb[indx].flags = flags;
	info->cr_cb_count = next;

	return indx;
    } else {
	return -1;
    }
}

// Note that we don't replace 'func', 'arg' and 'flags' atomically.
// Note that we ensure the context is not changed.
// Such atomicity is the responsibility of the caller.
//
// Not reentrant.
// Not exposed to user.
int
cri_do_replace(cri_info_t	*info,
	       cr_callback_id_t	id,
	       cr_callback_t	func,
	       void*		arg,
	       int		flags)
{
    int retval = -1;

    if ((flags & CRI_CONTEXT_MASK) == (id & CRI_CONTEXT_MASK)) {
	id &= ~CRI_CONTEXT_MASK;
	if ((id >= 0) && (id < info->cr_cb_count)) {
	    info->cr_cb[id].func  = func;
	    info->cr_cb[id].arg   = arg;
	    info->cr_cb[id].flags = flags;
	    retval = 0;
	} else {
	    /* id out-of-bounds */
	}
    } else {
	/* Context mismatch */
    }

    return retval;
}

cr_callback_id_t
cr_register_callback(cr_callback_t	func,
		     void*		arg,
		     int		flags)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    int context;
    int retval;

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    context = flags & CRI_CONTEXT_MASK;

    switch (context) {
    case CR_SIGNAL_CONTEXT:
	retval = cri_register_signal(info, func, arg, flags);
	break;
	
    case CR_THREAD_CONTEXT:
	retval = cri_register_thread(info, func, arg, flags);
	break;

    default:
	retval = -1;
    }

    return (retval >= 0) ? (retval | context) : retval;
}

// Not reentrant.
int
cr_replace_callback(cr_callback_id_t	id,
		    cr_callback_t	func,
		    void*		arg,
		    int			flags)
{
    cri_info_t *info = cri_info_location();    // thread-specific
    int retval;

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    switch (flags & CRI_CONTEXT_MASK) {
    case CR_SIGNAL_CONTEXT:
	retval = cri_replace_signal(info, id, func, arg, flags);
	break;
	
    case CR_THREAD_CONTEXT:
	retval = cri_replace_thread(info, id, func, arg, flags);
	break;

    default:
	retval = -1;
    }

    return retval;
}

// Not reentrant.
int
cr_replace_self(cr_callback_t	func,
		void*		arg,
		int		flags)
{
    cri_info_t *info = cri_info_location();	// thread-specific
    cr_callback_id_t id;

    if (!info || (cri_atomic_read(&info->cr_state) != CR_STATE_ACTIVE)) {
	// Not called from a callback!
	return -1;
    }

    /* form the full id from the index and context */
    id = info->run.id;
    id |= (info->cr_cb[id].flags & CRI_CONTEXT_MASK);

    return cri_do_replace(info, id, func, arg, flags);
}

int
cr_status(void)
{
    cri_info_t *info = cri_info_location();	// thread-specific

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    return do_state(info);
}

// XXX: Unused 'id' argument should be used for debugging.
void
cr_enter_cs(cr_client_id_t id)
{
    cri_info_t *info = cri_info_location();	// thread-specific

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    do_enter(info, id);
}

// XXX: Unused 'id' argument should be used for debugging.
void
cr_leave_cs(cr_client_id_t id)
{
    cri_info_t *info = cri_info_location();	// thread-specific

    if (!info) {
	CRI_ABORT("called w/o first calling cr_init()");
    }

    do_leave(info, id);
}

// Return address of the per-checkpoint/per-thread info
const struct cr_checkpoint_info * cr_get_checkpoint_info(void)
{
    cri_info_t *info = cri_info_location();		// thread-specific
    struct cr_checkpoint_info *result =  &(info->cr_checkpoint_info);

    if (result->dest == NULL) {
    	int rc = cri_syscall(CR_OP_HAND_DEST, (uintptr_t)info->path);
	if (rc != 0) {
	    CRI_ABORT("cri_syscall(CR_OP_HAND_DEST) returned w/ errno=%d",
		      errno);
	}
	result->dest = info->path;
    }

    return result;
}

// Return address of the per-restart/per-thread info
const struct cr_restart_info * cr_get_restart_info(void)
{
    cri_info_t *info = cri_info_location();		// thread-specific
    struct cr_restart_info *result =  &(info->cr_restart_info);

    result->requester = info->run.rc;	// Pid of cr_restart was returned from syscall

    if (result->src == NULL) {
    	int rc = cri_syscall(CR_OP_HAND_SRC, (uintptr_t)info->path);
	if (rc != 0) {
	    CRI_ABORT("cri_syscall(CR_OP_HAND_SRC) returned w/ errno=%d",
		      errno);
	}
	result->src = info->path;
    }
    return result;
}

// Return a client identifier used to invoke cr_{enter,leave}_cs().
// Also registers the handler on first call per thread.
// May later setup additional state.
cr_client_id_t cr_init(void)
{
    cri_info_t *info;
    int rc;

    // First verify that we have kernel support
    rc = cri_connect();
    if (rc < 0) {
	return rc;
    }

    info = cri_info_init();		// thread-specific
    if (!info->handler) {
	// Set the handler
	info->handler = &my_handler;

	// Register ourself with the kernel
	do {
	    rc = cri_syscall(CR_OP_HAND_PHASE2, (uintptr_t)&my_handler);
	} while ((rc < 0) && (errno == EAGAIN));
	if (rc != 0) {
	    CRI_ABORT("CR_OP_HAND_PHASE2 returned w/ errno = %d", errno);
	}
    }

    return info->next_id++;
}
